<?php

/**
 * CatiGood XML export PrestaShop module
 * v0.84 2011-07-25
 *
 * @author Marek Mauder <marek.mauder@catigood.cz>
 * @copyright Copyright (c) 2010-2011, CatiGood s.r.o.
 *
 * Change log:
 *   v0.84 (2011-07-25): fixed "hack attemp" error during export in PS 1.3.0.x
 *   v0.83 (2011-05-12): updated links to catigood.cz resources
 *   v0.82 (2011-01-24): compatibility with Presta 1.1
 *   v0.82 (2011-01-13): closed beta                       
 *   v0.80 (2010-12-23): internal testing version
 *
*/

/**
 * PrestaShop admin module for exporting products into Cati XML format.
 */
class CatiXML extends Module
{
    private $langId;
    private $content;
    private $feedPath;
    private $feedDir;
    private $baseUri;
    private $exportError;
    private $exportSuccess;
    private $link;        
    private $catFlat;
    private $attribGroups;

    const FeedDir = 'xml';
    const FeedFileName = 'catigood.xml';
    const CatSeparator = ' | ';
    const ModuleDownloadUri =  'http://data.catigood.cz/downloads/presta/catixml.zip';
    const VersionInfoUri = 'http://data.catigood.cz/downloads/presta/catixml.ver';

    public function __construct()
    {
        $this->name = 'catixml';

        parent::__construct();

        $this->page = basename(__FILE__, '.php');
        $this->tab = 'Tools';
        $this->version = 0.83;
        $this->displayName = $this->l('XML export pro CatiGood');
        $this->description = $this->l('Generuje XML export zboží pro službu CatiGood');
        $this->confirmUninstall = $this->l('Opravdu chcete odinstalovat modul XML export pro CatiGood?');

        $this->baseUri = 'http://' . $_SERVER['SERVER_NAME'] . __PS_BASE_URI__;
        $this->langId = Configuration::get('PS_LANG_DEFAULT');
        $this->feedDir = _PS_ROOT_DIR_ . '/' . self::FeedDir . '/';
        $this->feedPath = $this->feedDir . self::FeedFileName;
        $this->feedUri = $this->baseUri . self::FeedDir . '/' . self::FeedFileName;
        $this->link = new Link();
        $this->catFlat = array();
    }

    public function getContent()
    {
        global $cart;
        if (!isset($cart) OR !$cart->id) 
            $cart = new Cart(); // added as workaround for "hack attemp" eror in Tax::getApplicableTax in PS 1.3.0.x
        
        if (Tools::isSubmit('export'))
            $this->exportError = $this->doExport();

        $this->displayForm();

        return $this->content;
    }

    private function displayForm()
    {   
        // Show header with info and Cati contact
        $this->content .= '
            <h2 style="float: left; width: 700px;">Export zboží pro CatiGood</h2>
            <div style="float: right; width: 190px; display: inline-block;" class="right">
                <a href="http://catigood.cz" target="_blank" style="display: inline-block; margin: 10px;">
                    <img src="http://data.catigood.cz/images/logo/web.png" alt="CatiGood"/>
                </a>
                <br/>
                <a href="http://admin.catigood.cz" target="_blank" style="color: blue;" class="link">
                    Administrace F-Shopu
                </a>
                Technická podpora:<br/>
                <a style="color: blue;" class="link" href="&#109;&#097;&#105;&#108;&#116;&#111;&#058;&#x70;o&#x64;&#x70;or&#97;&#64;&#x61;dm&#105;&#110;.&#x63;a&#116;&#105;&#103;o&#x6f;&#100;&#46;&#x63;&#122;">
                    &#x70;o&#x64;&#x70;or&#97;&#64;&#x61;dm&#105;&#110;.&#x63;a&#116;&#105;&#103;o&#x6f;&#100;&#46;&#x63;&#122;
                </a>
            </div>
            <p style="float: left; width: 700px;">
                Níže uvedenou URL adresu vygenerovaného XML souboru vložte do
                admistrace Vašeho F-Shopu (záložka Katalog a potvrďte "Import zboží a uložit XML").
                Případné dotazy směřujte na naši technickou podporu. 
            </p>
            <div class="clear"></div>';

        if ($this->exportError)
        {
            // Print export error message if present
            $this->content .= '<p class="alert error">' . $this->exportError . '</p>';
        }
        else if (file_exists($this->feedPath))
        {
            // Report export success message if present
            if ($this->exportSuccess)
                $this->content .= '<p class="conf">' . $this->exportSuccess . '</p>';

            // Info about current XML file
            $date = date('j. n. Y H:i:s', filemtime($this->feedPath));
            $this->content .= '
                <p class="conf">
                    XML feed vygenerován dne ' . $date . ' na URL:<br/>
                    <a href="' . $this->feedUri . '" target="_blank" style="color: blue;" class="link">' . $this->feedUri . '</a>
                </p>';
        }
        else
        {
            // No XML file found
            $this->content .= '<p class="alert error">XML feed nebyl zatím vygenerován</p>';
        }

        // Check if there is a new version of this module available
        $versionInfo = @file_get_contents(self::VersionInfoUri);
        if ($versionInfo)
        {
            $versionInfo = explode('|', $versionInfo);
            $thisVer = (float)$this->version;
            $remoteVer = (float)$versionInfo[0];

            if ($remoteVer > $thisVer)
            {
                // Newer version detected
                $date = strtotime($versionInfo[1]);
                $this->content .= '<p class="warning">Existuje novější verze tohoto modulu
                    (' . $remoteVer . ' ze dne ' . date('j. n. Y', $date) . ') než
                    je nainstalována (' . $thisVer . '). Aktuální verzi si můžete stáhnout ze
                    <a href="' . self::ModuleDownloadUri . '" target="_blank" style="color: blue;"
                    class="link">serveru CatiGood</a>.</p>';
            }
        }

        $this->content .= '
            <form action="'. $REQUEST_URI.'" method="post">
                <input type="submit" name="export" value="Vygeneruj aktuální XML" style="padding: 10px; font-weight: bold; font-size: 120%;"/>
            </form>';
    }

    /**
     * Makes sure that given directory exists.     
     */
    private static function forceDirectories($dir, $mode = 0755)
    {
        if (is_dir($dir))
            return true;
        else
            return mkdir($dir, $mode, true);
    }

    /**
     * Store localized names of attributes and groups for later use
     * so we don't fetch strings from DB again and again for each product.
     */
    private function generateAttribCache()
    {
        $groups = AttributeGroup::getAttributesGroups($this->langId);

        foreach ($groups as $group)
        {
            $groupId = $group['id_attribute_group'];
            $attribs = AttributeGroup::getAttributes($this->langId, $groupId);

            foreach ($attribs as $attrib)
                 $group['attribs'][$attrib['id_attribute']] = $attrib['name'];

            $this->attribGroups[$groupId] = $group;
        }
    }

    private function prepareText($text, $stripTags = true)
    {
        $text = trim(html_entity_decode($text, ENT_COMPAT, 'UTF-8'));
        if ($stripTags)
            $text = strip_tags($text);
        return htmlspecialchars($text, ENT_COMPAT, 'UTF-8');        
    }

    /**
     * Generates flat category names (with all ancestors, e.g. "Components | GPUs | ATI")
     * for each node in given category tree. Works recursively starting at root node.
     */
    private function flattenCatTree(&$parent)
    {
        foreach ($parent['children'] as &$child)
        {
            if (isset($parent['displayName']))
                $child['displayName'] = $parent['displayName'] . self::CatSeparator . $child['name'];
            else
                $child['displayName'] = $child['name'];

            $this->catFlat[$child['id']] = $child['displayName'];
            $this->flattenCatTree($child);
        }
    }

    /**
     * Lists shop products with basic information. Only anctive items that
     * are in stock are listed.
     */
    private function getProductListing()
    {
        return Db::getInstance()->ExecuteS('
            SELECT p.`id_product`, pl.`name`
            FROM `'._DB_PREFIX_.'product` p
            LEFT JOIN `'._DB_PREFIX_.'product_lang` pl ON (p.`id_product` = pl.`id_product`)
            WHERE pl.`id_lang` = '.intval($this->langId).'
            AND p.active = 1 AND p.quantity > 0
            ORDER BY pl.`name`');
    }

    /**
     * Returns grouped attributes for given product (e.g. Color = {red, blue, yellow, ...}).
     * Each attribute can have price impact if it's defined in DB (only product attributes
     * generated by combination generator have entries in attrib impact table - manually created
     * combinations of attributes doesn't define impact and can be exported to Cati).
     */
    public function getProductAttribGroups($product)
    {
        // Get all product attributes for given product. There may be
        // many product attributes (id_product_attribute) based on same
        // attribute template (id_attribute).
        $attribs = Db::getInstance()->ExecuteS('
            SELECT ag.`id_attribute_group`, a.`id_attribute`,  pa.`id_product_attribute`               
            FROM `'._DB_PREFIX_.'product_attribute` pa
            LEFT JOIN `'._DB_PREFIX_.'product_attribute_combination` pac ON pac.`id_product_attribute` = pa.`id_product_attribute`
            LEFT JOIN `'._DB_PREFIX_.'attribute` a ON a.`id_attribute` = pac.`id_attribute`
            LEFT JOIN `'._DB_PREFIX_.'attribute_group` ag ON ag.`id_attribute_group` = a.`id_attribute_group`
            WHERE pa.`id_product` = '.intval($product->id).' AND pa.`quantity` > 0               
            ORDER BY pa.`id_product_attribute`');

        // Impacts define how each attribute will modify product price
        $impacts = Db::getInstance()->ExecuteS('
            SELECT ai.`id_attribute`, ai.`price`, ai.`weight`
	    FROM `'._DB_PREFIX_.'attribute_impact` ai
	    WHERE ai.`id_product` = '.intval($product->id));

        foreach ($impacts as $impact)
            $attribImpacts[$impact['id_attribute']] = $impact['price'];

        // Group attributes and assign price impacts if present
        foreach ($attribs as $key => $attrib)
        {
            $groupId = $attrib['id_attribute_group'];
            $attribId = $attrib['id_attribute'];

            $groups[$groupId][$attribId]['price'] = isset($attribImpacts[$attribId]) ? 
                $attribImpacts[$attribId] : 0;
        }

        return $groups;
    }

    /**
     * Check all images assigned to given product and 
     * store the link to full detail image for each of them.
     */
    private function saveImages($file, $product)
    {
        $imageArray = $product->getImages($product->id);
        $imageLinks = array();

        foreach ($imageArray as $img)
        {
            $filename = $product->id .'-'. $img['id_image'] .'.jpg';

            // Check if the image file exists in file system
            if (file_exists(_PS_PROD_IMG_DIR_ . $filename))
            {
                $uri = $this->baseUri . 'img/p/' . $filename;

                if ($img['cover'])
                    array_splice($imageLinks, 0, 0, $uri);   // cover image -> push to the index 0
                else
                    $imageLinks[] = $uri;
            }
        }

        foreach ($imageLinks as $link)
            fwrite($file, '    <IMGURL>' . $link . '</IMGURL>' . PHP_EOL);
    }

    /**
     * Store descriptive product features (e.g. date of manufacture,
     * dimensions, country of origin, ...).
     */
    private function saveProductFeature($file, $feat)
    {
        $lines = '  <PARAM>' . PHP_EOL .
                 '    <PARAM_NAME>' . $feat['name'] . '</PARAM_NAME>' . PHP_EOL .
                 '    <VAL>' . $feat['value'] . '</VAL>' . PHP_EOL .
                 '  </PARAM>' . PHP_EOL;
        fwrite($file, $lines);
    }

    /**
     * Save given product attribute group.
     */
    private function saveAttributeGroup($file, $groupId, $attribs)
    {
        $lines = '  <PARAM>' . PHP_EOL .
                 '    <PARAM_NAME>' . $this->attribGroups[$groupId]['public_name'] . '</PARAM_NAME>' . PHP_EOL;
        fwrite($file, $lines);

        foreach ($attribs as $id => $attrib)
        {
            $lines = '    <VAL';
            if ($attrib['price'] != 0)
                $lines .= ' PRICE="' . $attrib['price'] . '"';
            $lines .= '>' . $this->attribGroups[$groupId]['attribs'][$id] . '</VAL>' . PHP_EOL;
            fwrite($file, $lines);
        }

        fwrite($file, '  </PARAM>' . PHP_EOL);
    }

    /**
     * Save all product details to XML file.
     */
    private function saveProduct($file, $product)
    {
        $props = new stdClass();
        fwrite($file, '  <SHOPITEM>' . PHP_EOL);
  
        // Basic info
        $props->product = $this->prepareText($product->name);
        $props->url     = $this->link->getProductLink($product);
        $props->code    = $product->reference;
        $props->ean     = $product->ean13;
        $props->delivery_date = 0;

        // Prices
        $price = $product->getPrice(false, null);
        $tax = $product->tax_rate;
        if ($tax)
        {
            $props->price = number_format($price, 2, '.', '');
            $props->vat = $tax;            
        }
        else
            $props->price_vat = number_format($product->getPrice(true, null), 2, '.', '');

        // Category & Manufacturer
        $cats = Product::getIndexedCategories($product->id);        
        if (!empty($cats))
            $props->categorytext = $this->prepareText($this->catFlat[intval($cats[0]['id_category'])]);
        $props->manufacturer = $this->prepareText($product->manufacturer_name);        

        // Description
        $desc = $this->prepareText($product->description);
        $descShort = $this->prepareText($product->description_short);
        
        if ($descShort AND !$desc)
            $props->description = $descShort;
        else if (!$descShort AND $desc)
            $props->description = $desc;
        else
        {
            $props->description = $descShort;
            //$props->extdescription = $desc; disabled for now
        }        

        // Dump simple props to the file
        foreach ($props as $tag => $value)
        {
            $value = trim($value);
            if ($value !== null AND $value !== '')
                fwrite($file, '    <' . strtoupper($tag) . '>' . $value . '</' . strtoupper($tag) . '>' . PHP_EOL);
        }

        // Images
        $this->saveImages($file, $product);

        // Features - single value parameters
        $features = $product->getFrontFeatures($this->langId);
        foreach ($features as $key => $feat)
            $this->saveProductFeature($file, $feat);
        
        // Attributes - multi value parameters        
        $groups = $this->getProductAttribGroups($product);       
        foreach ($groups as $groupId => $attribs)
            $this->saveAttributeGroup($file, $groupId, $attribs);

        fwrite($file, '  </SHOPITEM>' . PHP_EOL);
    }

    /**
     * This is the actual export method.
     */
    private function doExport()
    {
        $time = microtime(true);

        // Make sure output directory exists
        if (!self::forceDirectories($this->feedDir))
            return 'Nepodařilo se vytvořit cílový adresář: ' . $this->feedDir;

        $file = fopen($this->feedPath, 'wb');
        if ($file === false)
            return 'Nepodařilo se vytvořit cílový soubor: ' . $this->feedPath;

        @set_time_limit(600);

        try
        {   
            fwrite($file, '<?xml version="1.0" encoding="UTF-8"?>' . PHP_EOL);
            fwrite($file, '<SHOP FORMAT="catigood">' . PHP_EOL);

            // Get category tree and prepare flattened category paths
            $rootCat = Category::getRootCategory($this->langId);
            $catTree = $rootCat->recurseLiteCategTree(10, 0, $this->langId);
            $this->flattenCatTree($catTree);
            // Generate product attrib cache
            $this->generateAttribCache();
            // Get list of all active products in stock
            $products = $this->getProductListing();

            foreach ($products as $pinfo)
            {
                // Get full info about current product and save it to the XML file
                $product = new Product($pinfo['id_product'], true, $this->langId);
                $this->saveProduct($file, $product);
                unset($product);
            }

            fwrite($file, '</SHOP>' . PHP_EOL);
            fclose($file);

            $time = microtime(true) - $time;
            $this->exportSuccess = sprintf('XML export úspěšně vytvořen - celkem %d produktů v %d kategoriích (%.2f sekund).',
                count($products), count($this->catFlat), $time);
        }
        catch (Exception $ex)
        {
            // Close and remove unfinished file
            fclose($file);
            unlink($this->feedPath);

            return sprintf('Neočekávaná chyba během exportu: %s (%s at line %d)',
                $ex->getMessage(), $ex->getFile(), $ex->getLine());
        }

        return null;
    }
}

?>
